#!/usr/bin/env python3

# sardinecake - manage VMs via OCI
#
# Copyright © 2024, IOhannes m zmölnig, forum::für::umläute
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


# parts if this code are:
# Copyright (c) 2021 Peter Odding
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import itertools
import json
import logging
import numbers
import os
import re

log = logging.getLogger("sardinecake.utils")

try:
    import magic

    def mimetype(filename: str) -> str | None:
        """guess the mimetype for a file (returns None if the filename cannot be guessed)"""
        try:
            import pathlib

            filename = pathlib.Path(filename).resolve()
        except:
            pass

        try:
            return magic.from_file(filename, mime=True)
        except OSError:
            return None

except ImportError:

    def mimetype(filename: str) -> str | None:
        """guess the mimetype for a file (returns None if the filename cannot be guessed)"""
        import subprocess

        try:
            p = subprocess.run(
                ["file", "-E", "-L", "--brief", "--mime-type", filename],
                stdout=subprocess.PIPE,
                check=True,
            )
            for line in p.stdout.splitlines():
                try:
                    # yikes, error messages go to stdout
                    if b" " in line:
                        continue
                    return line.decode()
                except:
                    continue
        except (OSError, subprocess.SubprocessError):
            return None
        return None


# © 2021, Peter Odding (expat)
# pilfered from https://github.com/xolox/python-humanfriendly/
def tokenize(text: str) -> list:
    """
    Tokenize a text into numbers and strings.

    :param text: The text to tokenize (a string).
    :returns: A list of strings and/or numbers.

    This function is used to implement robust tokenization of user input in
    functions like :func:`.parse_size()` and :func:`.parse_timespan()`. It
    automatically coerces integer and floating point numbers, ignores
    whitespace and knows how to separate numbers from strings even without
    whitespace. Some examples to make this more concrete:

    >>> from humanfriendly.text import tokenize
    >>> tokenize('42')
    [42]
    >>> tokenize('42MB')
    [42, 'MB']
    >>> tokenize('42.5MB')
    [42.5, 'MB']
    >>> tokenize('42.5 MB')
    [42.5, 'MB']
    """
    tokenized_input = []
    for token in re.split(r"(\d+(?:\.\d+)?)", text):
        token = token.strip()
        if re.match(r"\d+\.\d+", token):
            tokenized_input.append(float(token))
        elif token.isdigit():
            tokenized_input.append(int(token))
        elif token:
            tokenized_input.append(token)
    return tokenized_input


# © 2021, Peter Odding (expat)
# © 2024, IOhannes m zmölnig (AGPL)
# modified from https://github.com/xolox/python-humanfriendly/

size_units = {
    "k": 1024**1,
    "m": 1024**2,
    "g": 1024**3,
    "t": 1024**4,
    "p": 1024**5,
    "e": 1024**6,
    "z": 1024**7,
    "y": 1024**8,
    "kb": 1000**1,
    "mb": 1000**2,
    "gb": 1000**3,
    "tb": 1000**4,
    "pb": 1000**5,
    "eb": 1000**6,
    "zb": 1000**7,
    "yb": 1000**8,
    "kilobyte": 1000**1,
    "megabyte": 1000**2,
    "gigabyte": 1000**3,
    "terabyte": 1000**4,
    "petabyte": 1000**5,
    "exabyte": 1000**6,
    "zettabyte": 1000**7,
    "yottabyte": 1000**8,
    "kib": 1024**1,
    "mib": 1024**2,
    "gib": 1024**3,
    "tib": 1024**4,
    "pib": 1024**5,
    "eib": 1024**6,
    "zib": 1024**7,
    "yib": 1024**8,
    "kibibyte": 1024**1,
    "mebibyte": 1024**2,
    "gibibyte": 1024**3,
    "tebibyte": 1024**4,
    "pebibyte": 1024**5,
    "exbibyte": 1024**6,
    "zebibyte": 1024**7,
    "yobibyte": 1024**8,
}


def parse_size(size: str) -> numbers.Number:
    """parse a string with a a size-specifyer (such as 'MB' or 'G') into bytes
    the case of the specifier is ignored, as is whitespace.

    | specifier     | expands to  | value       |
    |---------------|-------------|-------------|
    | '7M'          | 7 MebiBytes | 7*1024*1024 |
    | '7MB'         | 7 MegaBytes | 7*1000*1000 |
    | '7 MiB'       | 7 MebiBytes | 7*1024*1024 |
    | '7 megabyte'  | 7 MegaBytes | 7*1000*1000 |
    | '7  mebiBYTE' | 7 MebiBytes | 7*1024*1024 |
    """
    tokens = tokenize(size)
    if not tokens:
        return None
    num = tokens[0]
    if not issubclass(type(num), numbers.Number):
        raise ValueError(f"Invalid number in size string {size!r}")
    if len(tokens) > 2:
        raise ValueError(f"Invalid size string {size!r}")
    factor = 1.0
    if len(tokens) == 2:
        factor = size_units.get(str(tokens[1]).lower())
        if not factor:
            raise ValueError(f"Invalid unit {tokens[1]} in {size!r}")
    return factor * num


def split_imagetag(path: str) -> (str, str | dict[str, str]):
    """splits a path with optional tag-decoration into a (path, tag) tuple.
    if the tag-decoration is a hashdigest, the returned <tag> is actually a
    <algo>: <digest> tuple.

    >>> split_imagetag('registry.example.org/sardines/debian:stable')
    ('registry.example.org/sardines/debian', 'stable')

    >>> split_imagetag('registry.example.org/sardines/debian')
    ('registry.example.org/sardines/debian', None)
    """
    try:
        path, digest = path.split("@", maxsplit=1)
        algo, digest = digest.split(":", maxsplit=1)
        return (path, (algo, digest))
    except ValueError:
        pass
    try:
        path, tag = path.split(":", maxsplit=1)
    except ValueError:
        tag = None
    return (path, tag)


def rmdirs(basedir: str, subdirs: list[str]) -> list[str]:
    """removes stacked subdirs, relative to <basedir>
    returns the full paths to the directories that have been removed

    >>> rmdirs('/tmp', ['', 'foo', 'bar'])
    ['/tmp/foo/bar', '/tmp/foo', '/tmp']
    """
    result = []
    for d in reversed(list(itertools.accumulate(subdirs, os.path.join))):
        xdir = os.path.join(basedir, d)
        try:
            os.rmdir(xdir)
            result.append(xdir)
        except OSError:
            return result
    return result


def checkSchema(data, schemaname: str):
    """checks whether <data> corresponds to JSON schema as described in the <schemaname> file."""
    import jsonschema

    schemafiles = [
        os.path.join("schema", f"{schemaname}.schema.json"),
        os.path.join(os.path.dirname(__file__), "schema", f"{schemaname}.schema.json"),
    ]
    schema = None
    for schemafile in schemafiles:
        try:
            log.debug(f"loading schema from {schemafile!r}")
            with open(schemafile) as f:
                schema = json.load(f)
                break
        except OSError:
            pass
    if not schema:
        return None
    try:
        jsonschema.validate(data, schema)
        return True
    except jsonschema.exceptions.ValidationError as e:
        log.debug(f"validating data against {schemaname} schema failed", exc_info=True)
        return False


class TryException(BaseException):
    """dummy exception that can be used to break out of a 'try'-clause"""

    pass


if __name__ == "__main__":

    def _show(size):
        print(f"{size!r} \t-> {parse_size(size)!r}")

    def _test():
        sizes = [
            "12",
            "1.2",
        ]
        units = "GB GiB kB kilobyte Gibibyte"
        for u in [None] + units.split():
            for s in sizes:
                if not u:
                    _show(s)
                else:
                    for f in ["%s%s", "%s %s", " %s\t%s "]:
                        _show(f % (s, u))

    _test()



#!/usr/bin/env python3

# sardinecake - manage VMs via OCI
#
# Copyright © 2024, IOhannes m zmölnig, forum::für::umläute
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


# login/logout to a registry

# ~/.docker/config.json
# containers-auth.json(5)
# https://github.com/docker/docker-credential-helpers/

import logging
import subprocess

log = logging.getLogger("sardinecake.auth")


def login(
    registry: str, username: str | None = None, password: str | None = None
) -> bool:
    """login to a registry"""
    cmd = ["skopeo", "login"]
    if username:
        cmd += ["--username", username]
    if password:
        cmd += ["--password-stdin"]
    cmd += [registry]
    if password:
        p = subprocess.Popen(cmd, stdin=subprocess.PIPE)
        p.communicate(password.encode())
    else:
        p = subprocess.run(cmd)

    return not p.returncode


def logout(registry: str | None):
    """logout from a registry
    if <registry> is 'None', logout from *all* registries
    """
    cmd = ["skopeo", "logout"]
    if registry is None:
        cmd.append("--all")
    else:
        cmd.append(registry)

    p = subprocess.run(cmd)
    return not p.returncode


def _main():
    import argparse

    parser = argparse.ArgumentParser(
        description="login to/logout from container registry on a specified server.",
    )

    p = parser.add_argument_group("action")
    p = p.add_mutually_exclusive_group()
    p.add_argument(
        "--login",
        action="store_true",
        default=True,
        help="login to the registry (DEFAULT)",
    )
    p.add_argument(
        "--logout",
        action="store_false",
        dest="login",
        help="logout from the registry",
    )

    p = parser.add_argument_group("credentials")
    p.add_argument(
        "--username",
        help="Username for registry",
    )
    p = p.add_mutually_exclusive_group()
    p.add_argument(
        "--password-stdin",
        action="store_true",
        help="Read the password from stdin",
    )
    p.add_argument(
        "--password",
        help="Password for registry",
    )

    p = parser.add_argument_group("protocol")
    p.add_argument(
        "--insecure",
        dest="secure",
        action="store_false",
        help="connect to the OCI registry via insecure HTTP protocol",
    )
    p.add_argument(
        "--secure",
        dest="secure",
        action="store_true",
        help="connect to the OCI registry via the secure HTTPS protocol (DEFAULT)",
    )

    p = parser.add_argument_group("credential validation")
    p.add_argument(
        "--no-validate",
        dest="validate",
        action="store_false",
        help="skip validation of the registry's credentials before logging-in",
    )

    p.add_argument(
        "--validate",
        dest="validate",
        action="store_true",
        help="validate the registry's credentials before logging-in (DEFAULT)",
    )

    parser.add_argument(
        "registry",
        help="registry to login to (e.g. 'ghcr.io')",
    )

    args = parser.parse_args()

    if args.login:
        x = login(args.registry)
    else:
        x = logout(args.registry)
    print(x)


if __name__ == "__main__":
    _main()



#!/usr/bin/env python3

# sardinecake - manage VMs via OCI
#
# Copyright © 2024, IOhannes m zmölnig, forum::für::umläute
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import logging
import subprocess

log = logging.getLogger("sardinecake.qcow")


class QCOW:
    def __init__(self, filename: str):
        try:
            self.getBackingChain = self._getBackingChain_pyqcow
        except AttributeError as e:
            pass
        self._filename = filename

    def getBackingChain(self, normalize=True) -> list[str]:
        """get the backing chain (recursively) for the QCOW2 image.
        returns the full backing chain, starting with ourself.
        if one backing-file could not be resolved, the list terminates with a None.
        on error, and an exception might be thrown
        """

        def actualNormalized(data: str) -> (str, str):
            """extract the actual resp. the normalized path, depending on the <normalize> value"""
            data = data.strip()
            if not data:
                return data
            data = [_.strip() for _ in data.split("(actual path: ", maxsplit=1)]
            if normalize and len(data) > 1:
                return data[1][:-1]
            return data[0]

        filename = self._filename

        p = subprocess.run(
            ["qemu-img", "info", "--backing-chain", filename],
            stdout=subprocess.PIPE,
            check=True,
            env={"LANG": "C"},
        )

        backingfiles = [
            actualNormalized(line.split(b":", maxsplit=1)[1].decode())
            for line in p.stdout.splitlines()
            if line.startswith(b"backing file:")
        ]
        return [filename] + backingfiles

    try:
        import pyqcow

        def _getBackingChain_pyqcow(self, normalize=True, _cwd=None):
            """set help(QCOW.getBackingChain) for a description"""
            import pyqcow

            filename = self._filename
            fullname = os.path.join(_cwd or "", filename)

            if normalize:
                result = [fullname]
            else:
                result = [filename]

            with open(fullname, "rb") as f:
                q = pyqcow.file()
                q.open_file_object(f)
                backingfile = q.get_backing_filename()
            if not backingfile:
                return result
            dirname = os.path.dirname(fullname)
            try:
                bfiles = (
                    QCOW(backingfile).getBackingChain(normalize=normalize, _cwd=dirname)
                    or []
                )
            except OSError:
                bfiles = [None]
            return result + bfiles

        del pyqcow

    except ImportError:
        pass

    def commit(self) -> list[str]:
        """commit any changes from the current file into it's backing-file (if any).
        this basically merges changes to the backing-file, and removes them from *this*file.

        returns the backing-chain if the given file.
        only the first item in the list is modified!

        e.g. snapshot.qcow2 -> child.qcow2 -> master.qcow2

        # merge all changes from 'snapshot.qcow2' into 'child.qcow2'
        # ('master.qcow2' is left as is)
        >>> QCOW("snapshot.qcow2").commit()
        ['child.qcow2', 'master.qcow2']

        # merge all changes from 'child.qcow2' into 'master.qcow2'
        # TODO: does this invalidate 'snapshot.qcow2'?
        >>> QCOW("child.qcow2").commit()
        ['master.qcow2']

        >>> QCOW("master.qcow2").commit()
        []
        """

        backingchain = self.getBackingChain()
        filename = self._filename

        if not backingchain:
            return None
        p = subprocess.run(
            ["qemu-img", "commit", filename],
            check=True,
        )
        return backingchain[1:]

    def create(self, size: int | None = None, backing_file: str | None = None):
        """create a new QCOW2 image.
        either <size> xor <backing_file> must be present.
        """
        if bool(size) == bool(backing_file):
            raise ValueError(
                f"You have to specify either size ({size}) OR backing_file ({backing_file})"
            )
        cmd = ["qemu-img", "create", "-f", "qcow2"]
        if backing_file:
            cmd += ["-F", "qcow2", "-b", backing_file]

        cmd.append(self._filename)

        if size:
            cmd.append(str(size))

        log.debug(cmd)

        args = {}
        if log.getEffectiveLevel() >= logging.INFO:
            args["stdout"] = subprocess.DEVNULL
            args["stderr"] = subprocess.DEVNULL
        else:
            args["stdout"] = subprocess.sys.stderr

        p = subprocess.run(
            cmd,
            **args,
        )

        if p.returncode:
            return False

        return True


def _main():
    import argparse

    parser = argparse.ArgumentParser(
        description="manipulate QCOW2 images",
    )
    subparsers = parser.add_subparsers(required=True)

    p_check = subparsers.add_parser(
        "backingchain", help="Get the backing chain of the QCOW2 files"
    )
    p_check.set_defaults(fun="getBackingChain")
    p_check.add_argument(
        "--no-normalize",
        action="store_false",
        dest="normalize",
        help="Output path names as stored within the qcow file(s).",
    )
    p_check.add_argument(
        "--normalize",
        action="store_true",
        default=True,
        help="Output normalized path names (DEFAULT)",
    )

    p_merge = subparsers.add_parser(
        "commit",
        help="Merge the changes from the QCOW2 file into its (1st level) backing file",
    )
    p_merge.set_defaults(fun="commit")

    for p in [p_check, p_merge]:
        p.add_argument("file", nargs="+", help="qcow2 file(s) to work on")

    args = parser.parse_args()

    kwargs = {}
    try:
        kwargs["normalize"] = args.normalize
    except AttributeError:
        pass
    print(f"applying {args.fun!r}")

    for filename in args.file:
        qcow = QCOW(filename)

        fun = getattr(qcow, args.fun)
        try:
            chain = fun(**kwargs)
        except Exception as e:
            log.exception(f"couldn't do {args.fun!r}")
            chain = e
        print(f"{filename} -> {chain}")


def getBackingChain(disk: str) -> list[str]:
    "gets backing chain if the disk (including the <disk> as first item)"
    try:
        return QCOW(disk).getBackingChain()
    except:
        pass
    return [disk]


if __name__ == "__main__":
    _main()
